@@ -32,6 +32,10 @@ LOGIN_URL = "/auth/login.json"
 VERIFY_URL = "/auth/verify.json"
 
 
+class PassboltValidationError(Exception):
+    pass
+
+
 class APIClient:
 
     def __init__(self, config: Optional[str] = None, config_path: Optional[str] = None, new_keys: bool = False,
@@ -52,7 +56,7 @@ class APIClient:
         if not self.config["PASSBOLT"]["SERVER"]:
             raise ValueError("Missing value for SERVER in config.ini")
 
-        self.server_url = self.config["PASSBOLT"]["SERVER"]
+        self.server_url = self.config["PASSBOLT"]["SERVER"].rstrip("/")
         self.user_fingerprint = self.config["PASSBOLT"]["USER_FINGERPRINT"].upper()
         self.gpg = gnupg.GPG()
         if delete_old_keys:
@@ -105,11 +109,15 @@ class APIClient:
                 "user_token_result": token
             },
         })
+        self._get_csrf_token()
 
-    def encrypt(self, text):
+    def _get_csrf_token(self):
+        self.get("/users/me.json", return_response_object=True)  # Fetches the X-CSRF-Token header for future requests
+
+    def encrypt(self, text, recipients=None):
         return str(self.gpg.encrypt(
             data=text,
-            recipients=self.gpg_fingerprint,
+            recipients=recipients or self.gpg_fingerprint,
             always_trust=True
         ))
 
@@ -128,133 +136,10 @@ class APIClient:
         r = self.requests_session.get(self.server_url + VERIFY_URL)
         return r.json()["body"]["fingerprint"], r.json()["body"]["keydata"]
 
-    def get(self, url):
-        r = self.requests_session.get(self.server_url + url)
-        return r.json()
-
-    def post(self, url, data):
-        r = self.requests_session.post(self.server_url + url, json=data, headers=self.get_headers())
-        return r.json()
-
-    def put(self, url, data):
-        r = self.requests_session.put(self.server_url + url, json=data, headers=self.get_headers())
-        return r.json()
-
     def delete(self, url):
         r = self.requests_session.delete(self.server_url + url, headers=self.get_headers())
         return r.json()
 
-    def close_session(self):
-        self.requests_session.close()
-
-
-class PassboltAPI(APIClient):
-    """Adding a convenience method for getting resources.
-
-    Design Principle: All passbolt aware public methods must accept or output one of PassboltTupleTypes"""
-
-    def iterate_resources(self, params: Optional[dict] = None):
-        params = params or {}
-        url_params = urllib.parse.urlencode(params)
-        if url_params:
-            url_params = "?" + url_params
-        response = self.get('/resources.json' + url_params)
-        assert "body" in response.keys(), f"Key 'body' not found in response keys: {response.keys()}"
-        resources = response["body"]
-        for resource in resources:
-            yield resource
-
-    def list_resources(self, folder_id: Union[None, PassboltFolderIdType] = None):
-        params = {
-            **({"filter[has-id][]": folder_id} if folder_id else {}),
-            "contain[children_resources]": True,
-        }
-        url_params = urllib.parse.urlencode(params)
-        if url_params:
-            url_params = "?" + url_params
-        response = self.get('/folders.json' + url_params)
-        assert "body" in response.keys(), f"Key 'body' not found in response keys: {response.keys()}"
-        response = response["body"][0]
-        assert "children_resources" in response.keys(), f"Key 'body[].children_resources' not found in response " \
-                                                        f"keys: {response.keys()} "
-        return constructor(PassboltResourceTuple)(response["children_resources"])
-
-    def get_secret(self, resource: PassboltResourceTuple) -> PassboltSecretTuple:
-        response = self.get(f"/secrets/resource/{resource.id}.json")
-        assert "body" in response.keys(), f"Key 'body' not found in response keys: {response.keys()}"
-        return PassboltSecretTuple(**response["body"])
-
-    def update_secret(self, resource: PassboltResourceTuple, new_secret):
-        return self.put(f"/resources/{resource.id}.json", {
-            "secrets": new_secret
-        }, return_response_object=True)
-
-    def list_users(self, can_access: Union[None, PassboltResourceTuple, PassboltFolderTuple] = None) \
-            -> List[PassboltUserTuple]:
-        if can_access is not None:
-            params = {"filter[has-access]": can_access.id}
-        else:
-            params = {}
-        response = self.get(f"/users.json", params=params)
-        assert "body" in response.keys(), f"Key 'body' not found in response keys: {response.keys()}"
-
-        response["body"]: List[Mapping]
-
-        return constructor(
-            PassboltUserTuple,
-            subconstructors={"gpgkey": constructor(PassboltOpenPgpKeyTuple)},
-        )(response["body"])
-
-    def import_public_keys(self, trustlevel='TRUST_FULLY'):
-        # get all users
-        users = self.list_users()
-        for user in users:
-            self.gpg.import_keys(user.gpgkey.armored_key)
-            self.gpg.trust_keys(user.gpgkey.fingerprint, trustlevel)
-
-    def describe_folder(self, folder_id):
-        return self.get(f"/folders/{folder_id}.json")
-
-    def create_resource(self, name: str, username: str, password: str, uri: Optional[str] = None,
-                        description: Optional[str] = None, folder: Optional[str] = None):
-        """Creates a new resource on passbolt and shares it with the provided group and folder"""
-        r = self.post("/resources.json", {
-            "name": name,
-            "username": username,
-            "description": description or "",
-            "uri": uri or "",
-            "secrets": [
-                {
-                    "data": self.encrypt(password)
-                }
-            ],
-            **(folder or {"folder_parent_id": folder}),
-        }, return_response_object=True)
-        r.raise_for_status()
-        r_json = r.json()
-        new_resource_id = r_json.get("body", {}).get("id")
-        if new_resource_id is None:
-            raise ValueError(f"Unexpected resource creation: {r_json}")
-        if folder:
-            self.move_resource_to_folder(new_resource_id, folder)
-        return r_json
-
-    def move_resource_to_folder(self, resource_id, folder_id):
-        r = self.post(f"/move/resource/{resource_id}.json", {"folder_parent_id": folder_id},
-                      return_response_object=True)
-        r.raise_for_status()
-        return r.json()
-
-    def encrypt(self, text, recipients=None):
-        if recipients is None:
-            return super().encrypt(text)
-        else:
-            return str(self.gpg.encrypt(
-                data=text,
-                recipients=recipients,
-                always_trust=True
-            ))
-
     def get(self, url, return_response_object=False, **kwargs):
         r = self.requests_session.get(self.server_url + url, headers=self.get_headers(), **kwargs)
         if return_response_object:
@@ -272,3 +157,203 @@ class PassboltAPI(APIClient):
         if return_response_object:
             return r
         return r.json()
+
+    def close_session(self):
+        self.requests_session.close()
+
+
+class PassboltAPI(APIClient):
+    """Adding a convenience method for getting resources.
+
+    Design Principle: All passbolt aware public methods must accept or output one of PassboltTupleTypes"""
+
+    def _encrypt_secrets(self, secret_text: str, recipients: List[PassboltUserTuple]) -> List[Mapping]:
+        return [
+            {
+                "user_id": user.id,
+                "data": self.encrypt(secret_text, user.gpgkey.fingerprint)
+            }
+            for user in recipients
+        ]
+
+    def iterate_resources(self, params: Optional[dict] = None):
+        params = params or {}
+        url_params = urllib.parse.urlencode(params)
+        if url_params:
+            url_params = "?" + url_params
+        response = self.get('/resources.json' + url_params)
+        assert "body" in response.keys(), f"Key 'body' not found in response keys: {response.keys()}"
+        resources = response["body"]
+        for resource in resources:
+            yield resource
+
+    def list_resources(self, folder_id: Optional[PassboltFolderIdType] = None):
+        params = {
+            **({"filter[has-id][]": folder_id} if folder_id else {}),
+            "contain[children_resources]": True,
+        }
+        url_params = urllib.parse.urlencode(params)
+        if url_params:
+            url_params = "?" + url_params
+        response = self.get('/folders.json' + url_params)
+        assert "body" in response.keys(), f"Key 'body' not found in response keys: {response.keys()}"
+        response = response["body"][0]
+        assert "children_resources" in response.keys(), f"Key 'body[].children_resources' not found in response " \
+                                                        f"keys: {response.keys()} "
+        return constructor(PassboltResourceTuple)(response["children_resources"])
+
+    def get_secret(self, resource_id: PassboltResourceIdType) -> PassboltSecretTuple:
+        response = self.get(f"/secrets/resource/{resource_id}.json")
+        assert "body" in response.keys(), f"Key 'body' not found in response keys: {response.keys()}"
+        return PassboltSecretTuple(**response["body"])
+
+    def update_secret(self, resource_id: PassboltResourceIdType, new_secret):
+        return self.put(f"/resources/{resource_id}.json", {
+            "secrets": new_secret
+        }, return_response_object=True)
+
+    def list_users(self, resource_or_folder_id: Union[None, PassboltResourceIdType, PassboltFolderIdType] = None, force_list=True) \
+            -> List[PassboltUserTuple]:
+        if resource_or_folder_id is None:
+            params = {}
+        else:
+            params = {"filter[has-access]": resource_or_folder_id}
+        params["contain[permission]"] = True
+        response = self.get(f"/users.json", params=params)
+        assert "body" in response.keys(), f"Key 'body' not found in response keys: {response.keys()}"
+        response = response["body"]
+        users = constructor(
+            PassboltUserTuple,
+            subconstructors={
+                "gpgkey": constructor(PassboltOpenPgpKeyTuple),
+            },
+        )(response)
+        if isinstance(users, PassboltUserTuple) and force_list:
+            return [users]
+        return users
+
+    def import_public_keys(self, trustlevel='TRUST_FULLY'):
+        # get all users
+        users = self.list_users()
+        for user in users:
+            self.gpg.import_keys(user.gpgkey.armored_key)
+            self.gpg.trust_keys(user.gpgkey.fingerprint, trustlevel)
+
+    def describe_folder(self, folder_id: PassboltFolderIdType):
+        return self.get(f"/folders/{folder_id}.json")
+
+    def create_resource(self, name: str, password: str,
+                        username: str = "",
+                        description: str = "",
+                        uri: str = "",
+                        resource_type_id: Optional[PassboltResourceTypeIdType] = None,
+                        folder_id: Optional[PassboltFolderIdType] = None):
+        """Creates a new resource on passbolt and shares it with the provided folder recipients"""
+        if not name:
+            raise PassboltValidationError(f"Name cannot be None or empty -- {name}!")
+        if not password:
+            raise PassboltValidationError(f"Password cannot be None or empty -- {password}!")
+
+        r_create = self.post("/resources.json", {
+            "name": name,
+            "username": username,
+            "description": description,
+            "uri": uri,
+            **({"resource_type_id": resource_type_id} if resource_type_id else {}),
+            "secrets": [
+                {
+                    "data": self.encrypt(password)
+                }
+            ],
+        }, return_response_object=True)
+        r_create.raise_for_status()
+        resource = constructor(PassboltResourceTuple)(r_create.json()["body"])
+        if folder_id:
+            folder = self.read_folder(folder_id)
+            # get users with access to folder
+            lookup_users: Mapping[PassboltUserIdType, PassboltUserTuple] = {
+                user.id: user for user in self.list_users(resource_or_folder_id=folder_id)
+            }
+            # simulate sharing with folder perms
+            share_payload = {
+                "permissions": [
+                    {
+                        "is_new": True,
+                        "user": lookup_users.get(perm.aro_foreign_key) and lookup_users.get(
+                            perm.aro_foreign_key).username,
+                        **{k: v for k, v in perm._asdict().items() if k != "id"},
+                    } for perm in folder.permissions
+                    if perm.aro == "User" and (
+                            lookup_users.get(perm.aro_foreign_key).gpgkey.fingerprint
+                            != self.user_fingerprint
+                    )
+                ],
+                "secrets": self._encrypt_secrets(password, lookup_users.values())
+            }
+            r_simulate = self.post(f"/share/simulate/resource/{resource.id}.json",
+                                   share_payload, return_response_object=True)
+            r_simulate.raise_for_status()
+
+            r_share = self.put(f"/share/resource/{resource.id}.json", share_payload, return_response_object=True)
+            r_share.raise_for_status()
+
+            self.move_resource_to_folder(resource_id=resource.id, folder_id=folder_id)
+        return r_create
+
+    def update_resource(self,
+                        resource_id: PassboltResourceIdType,
+                        name: Optional[str] = None,
+                        username: Optional[str] = None,
+                        description: Optional[str] = None,
+                        uri: Optional[str] = None,
+                        resource_type_id: Optional[PassboltResourceTypeIdType] = None,
+                        password: Optional[str] = None):
+        payload = {
+            "name": name,
+            "username": username,
+            "description": description,
+            "uri": uri,
+            "resource_type_id": resource_type_id,
+        }
+        if name is None:
+            payload.pop("name")
+        if username is None:
+            payload.pop("username")
+        if description is None:
+            payload.pop("description")
+        if uri is None:
+            payload.pop("uri")
+        if resource_type_id is None:
+            payload.pop("resource_type_id")
+
+        if password is not None:
+            assert isinstance(password, str), f"password has to be a string object -- {password}"
+            recipients = self.list_users(resource_or_folder_id=resource_id)
+
+            payload["secrets"] = self._encrypt_secrets(password, recipients=recipients)
+        if payload:
+            r = self.put(f"/resources/{resource_id}.json", payload, return_response_object=True)
+            r.raise_for_status()
+            return r
+
+    def move_resource_to_folder(self, resource_id: PassboltResourceIdType, folder_id: PassboltFolderIdType):
+        r = self.post(f"/move/resource/{resource_id}.json", {"folder_parent_id": folder_id},
+                      return_response_object=True)
+        r.raise_for_status()
+        return r.json()
+
+    def read_folder(self, folder_id: PassboltFolderIdType) -> PassboltFolderTuple:
+        response = self.get(f"/folders/{folder_id}.json", params={"contain[permissions]": True},
+                            return_response_object=True)
+        response.raise_for_status()
+        response = response.json()
+        return constructor(PassboltFolderTuple,
+                           subconstructors={
+                               "permissions": constructor(PassboltPermissionTuple)
+                           })(response['body'])
+
+    def read_resource(self, resource_id: PassboltResourceIdType) -> PassboltResourceTuple:
+        response = self.get(f"/resources/{resource_id}.json", return_response_object=True)
+        response.raise_for_status()
+        response = response.json()["body"]
+        return constructor(PassboltResourceTuple)(response)
